Program TMap;
{$N+,E+}

Uses
  Crt,
  Types,
  Inertia,
  Bufio;

{$I VERSION.INC}

Const
  PLANAR = 0;
  CYLINDRICAL = 1;
  SPHERICAL = 2;
  XX = 0;
  XY = 1;
  XZ = 2;
  YX = 3;
  YY = 4;
  YZ = 5;
  ZX = 6;
  ZY = 7;
  ZZ = 8;
  MappingStyle : Integer = SPHERICAL;
  MappingAxis  : Integer = XY;

Var
  Obj : VectorObject;
  InFile, OutFile : String;
  FitTexture : Boolean;

Function UpStr(S : String) : String;

Var
  Temp : String;
  C : Integer;

Begin
  Temp := '';
  For C := 1 to Length(S) do
    Temp := Temp + UpCase(S[C]);
  UpStr := Temp;
End;

Procedure Help;

Begin
  Writeln('Usage: TMAP {file} {options}');
  Writeln('Options: (preceded by ''/'' or ''-'')');
  Writeln;
  Writeln('                 [File Specifiers]');
  Writeln('{file} specifies <infile> [outfile]');
  Writeln('   <infile>: input GVO file (.GVO is default extention)');
  Writeln('   <outfile>: optional output GVO file (.GVO is default extention)');
  Writeln;
  Writeln('                 [Mapping Options] ');
  Writeln(' /s   spherical texture coordinate generation (default)');
  Writeln(' /p   planar texture coordinate generation');
  Writeln(' /c   cylindrical texture coordinate generation');
  Writeln(' /f   stretch texture coordinates to fill full texture space');
  Writeln;
  Writeln('                 [Axis Options]');
  Writeln(' /a<##> choose axis for mapping. ## can be any combination of ''X'', ''Y'' and ''Z''');
  Writeln('        defaults to ''XX''');
  Writeln;
  Writeln('                 [Help]');
  Writeln(' [/h | /?]  displays this help.');
  Halt(0);
End;

Procedure Clip(Var i : Integer; Low, High : Integer);

Begin
  If (i < Low)
    Then i := Low
    Else If (i > High)
      Then i := High;
End;

Function GetMapping(S : String) : Integer;

Var
  Temp : Integer;
  Error : Integer;

Begin
  If (Length(S) = 2)
    Then Begin
      GetMapping := -1;
      Exit;
    End;
  If (Length(S) < 4)  { check for just 1 axis specification }
    Then Begin
      Case (S[3]) of
        'X' : GetMapping := XX;
        'Y' : GetMapping := YY;
        'Z' : GetMapping := ZZ;
      End;
    End
    Else Begin
      Case (S[3]) of
        'X' : Begin
                Case S[4] of
                  'X' : GetMapping := XX;
                  'Y' : GetMapping := XY;
                  'Z' : GetMapping := XZ;
                  Else  GetMapping := -1;
                End;
              End;
        'Y' : Begin
                Case S[4] of
                  'X' : GetMapping := YX;
                  'Y' : GetMapping := YY;
                  'Z' : GetMapping := YZ;
                  Else  GetMapping := -1;
                End;
              End;
        'Z' : Begin
                Case S[4] of
                  'X' : GetMapping := ZX;
                  'Y' : GetMapping := ZY;
                  'Z' : GetMapping := ZZ;
                  Else  GetMapping := -1;
                End;
              End;
        Else  GetMapping := -1;
      End;
    End;
End;


Procedure ParseCommandLine;

Var
  i : Integer;
  Temp : String;
  InFlag : Boolean;
  OutFlag : Boolean;

Begin
  Writeln('TMap - Texture Map v',VERSION_MAJOR,'.',VERSION_MINOR,'                     Alex Chalfin 8/2/96');
  Writeln;
  InFlag := False;
  OutFlag := False;
  FitTexture := False;
  If ParamCount = 0 Then Help;
  For i := 1 to ParamCount do
    Begin
      Temp := UpStr(ParamStr(i));
      If (Temp[1] = '-') or (Temp[1] = '/')
        Then Begin
          If Temp[2] = 'P' Then MappingStyle := PLANAR;
          If Temp[2] = 'C' Then MappingStyle := CYLINDRICAL;
          If Temp[2] = 'S' Then MappingStyle := SPHERICAL;
          If Temp[2] = 'A' Then MappingAxis := GetMapping(Temp);
          If Temp[2] = 'F' Then FitTexture := True;
        End
        Else Begin
          If Not(InFlag)
            Then Begin
              InFile := Temp;
              InFlag := True;
            End
            Else Begin
              OutFile := Temp;
              OutFlag := True;
            End;
        End;
    End;
  If Not(InFlag)
    Then Begin
      Writeln('No input file specified!');
      Halt(0);
    End;
  If Pos('.', InFile) = 0
    Then InFile := InFile + '.GVO';
  If Not(OutFlag)
    Then Begin
      OutFile := InFile;
      While(OutFile[Length(OutFile)] <> '.') do
        Delete(OutFile, Length(OutFile), 1);
      OutFile := OutFile + 'GVO';
    End
    Else Begin
      If Pos('.', OutFile) = 0
        Then OutFile := OutFile + '.GVO';
    End;
  If (MappingAxis = -1)
    Then Begin
      Writeln('Mapping axis selection is invalid.');
      Halt(0);
    End;
End;

Procedure InitObject;

Begin
  If LoadGVO(Obj, InFile, 0) <> 0
    Then Begin
      Writeln('Error loading ',InFile);
      Halt(0);
    End;
  If (Obj.Texture_Coords = Nil)
    Then Getmem(Obj.Texture_Coords, Obj.Num_Polygons * Sizeof(TextureCoord));
End;

Function WriteGVOFile(Filename : String) : Integer;

Var
  Diskfile : bFile;
  Flag : Word;
  i : Integer;
  t : Byte;
  Temp : Integer;
  FBuffer : Array[0..1023] of Byte;

Begin
  FAssign(Diskfile, Filename);
  FSetBuffer(Diskfile, FBuffer, Sizeof(FBuffer));
  FReWrite(Diskfile);
  If IOResult <> 0
    Then Begin
      WriteGVOFile := 1;
      Exit;
    End;
  t := Ord('G');
  FWrite(Diskfile, t, 1);
  t := Ord('V');
  FWrite(Diskfile, t, 1);
  t := Ord('O');
  FWrite(Diskfile, t, 1);
  Flag := 31;   { 0000000000011111b }
  FWrite(Diskfile, Flag, Sizeof(Flag));
  FWrite(Diskfile, Obj.Num_Verticies, Sizeof(Integer));
  FWrite(Diskfile, Obj.Object_Definition^, Obj.Num_Verticies*Sizeof(ShortVertex3d));
  FWrite(Diskfile, Obj.Num_Polygons, Sizeof(Integer));
  For i := 0 to (Obj.Num_Polygons - 1) do
    Begin
      FWrite(Diskfile, Obj.Polygons^[i].A, Sizeof(Integer));
      FWrite(Diskfile, Obj.Polygons^[i].B, Sizeof(Integer));
      FWrite(Diskfile, Obj.Polygons^[i].C, Sizeof(Integer));
    End;
  
  FWrite(Diskfile, Obj.Polygon_Normals^, Obj.Num_Polygons*Sizeof(ShortVertex3d));
  FWrite(Diskfile, Obj.Vertex_Normals^, Obj.Num_Verticies*Sizeof(ShortVertex3d));

  For i := 0 to (Obj.Num_Polygons - 1) do
    Begin
      FWrite(Diskfile, Obj.Texture_Coords^[i].UV1, Sizeof(UVCoord));
      FWrite(Diskfile, Obj.Texture_Coords^[i].UV2, Sizeof(UVCoord));
      FWrite(Diskfile, Obj.Texture_Coords^[i].UV3, Sizeof(UVCoord));
    End;
  Flush(Diskfile);
  FClose(Diskfile);
End;

Procedure AddUVCoord(u, v : Byte; vert : Integer);

Var
  i : Integer;

Begin
  For i := 0 to (Obj.Num_Polygons - 1) do
    Begin
      If (Obj.Polygons^[i].A = vert)
        Then Begin
          Obj.Texture_Coords^[i].UV1.u := u;
          Obj.Texture_Coords^[i].UV1.v := v;
        End;
      If (Obj.Polygons^[i].B = vert)
        Then Begin
          Obj.Texture_Coords^[i].UV2.u := u;
          Obj.Texture_Coords^[i].UV2.v := v;
        End;
      If (Obj.Polygons^[i].C = vert)
        Then Begin
          Obj.Texture_Coords^[i].UV3.u := u;
          Obj.Texture_Coords^[i].UV3.v := v;
        End;
    End;
End;

Procedure PlanarMapping;

Var
  i : Integer;
  xMin, xMax : Longint;
  yMin, yMax : Longint;
  zMin, zMax : Longint;
  xSize, ySize, zSize : Longint;
  xScale, yScale, zScale : Real;
  tempu, tempv : Integer;

Begin
  xMin := MaxInt;
  xMax := -MaxInt;
  yMin := MaxInt;
  yMax := -MaxInt;
  zMin := MaxInt;
  zMax := -MaxInt;
  For i := 0 to (Obj.Num_Verticies - 1) do
    Begin
      If (Obj.Object_Definition^[i].x < xMin)
        Then xMin := Obj.Object_Definition^[i].x;
      If (Obj.Object_Definition^[i].x > xMax)
        Then xMax := Obj.Object_Definition^[i].x;
      If (Obj.Object_Definition^[i].y < yMin)
        Then yMin := Obj.Object_Definition^[i].y;
      If (Obj.Object_Definition^[i].y > yMax)
        Then yMax := Obj.Object_Definition^[i].y;
      If (Obj.Object_Definition^[i].z < zMin)
        Then zMin := Obj.Object_Definition^[i].z;
      If (Obj.Object_Definition^[i].z > zMax)
        Then zMax := Obj.Object_Definition^[i].z;
    End;
  xSize := Abs(xMax) + Abs(xMin);
  ySize := Abs(yMax) + Abs(yMin);
  zSize := Abs(zMax) + Abs(zMin);
  If FitTexture
    Then Begin
      xScale := 256/xSize;
      yScale := 256/ySize;
      zScale := 256/zSize;
    End
    Else Begin
      If (xSize > ySize)
        Then Begin
          If (xSize > zSize)
            Then xScale := 256/xSize
            Else xScale := 256/zSize;
        End
        Else Begin
          If (ySize > zSize)
            Then xScale := 256/ySize
            Else xScale := 256/zSize;
        End;
      yScale := xScale;
      zScale := xScale;
    End;
  For i := 0 to (Obj.Num_Verticies - 1) do
    Begin
      Case MappingAxis of
        XX : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        XY : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        XZ : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;
        YX : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        YY : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        YZ : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;
        ZX : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        ZY : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        ZZ : Begin
               tempu := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;
      End;
      Clip(tempu, 0, 255);
      Clip(tempv, 0, 255);
      AddUVCoord(tempu, tempv, i);
    End;
End;

Procedure SphericalMapping;

Var
  tempu, tempv : Integer;
  i : Integer;

Begin
  For i := 0 to (Obj.Num_Verticies - 1) do
    Begin
      Case MappingAxis of
        XX : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
             End;
        XY : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
             End;
        XZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
             End;
        YX : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
             End;
        YY : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
             End;
        YZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
             End;
        ZX : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
             End;
        ZY : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
             End;
        ZZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
             End;
      End;
      Clip(tempu, 0, 255);
      Clip(tempv, 0, 255);
      AddUVCoord(tempu, tempv, i);
    End;
End;

Procedure CylindricalMapping;

Var
  i : Integer;
  xMin, xMax : Longint;
  yMin, yMax : Longint;
  zMin, zMax : Longint;
  xSize, ySize, zSize : Longint;
  xScale, yScale, zScale : Real;
  tempu, tempv : Integer;

Begin
  xMin := MaxInt;
  xMax := -MaxInt;
  yMin := MaxInt;
  yMax := -MaxInt;
  zMin := MaxInt;
  zMax := -MaxInt;
  For i := 0 to (Obj.Num_Verticies - 1) do
    Begin
      If (Obj.Object_Definition^[i].x < xMin)
        Then xMin := Obj.Object_Definition^[i].x;
      If (Obj.Object_Definition^[i].x > xMax)
        Then xMax := Obj.Object_Definition^[i].x;
      If (Obj.Object_Definition^[i].y < yMin)
        Then yMin := Obj.Object_Definition^[i].y;
      If (Obj.Object_Definition^[i].y > yMax)
        Then yMax := Obj.Object_Definition^[i].y;
      If (Obj.Object_Definition^[i].z < zMin)
        Then zMin := Obj.Object_Definition^[i].z;
      If (Obj.Object_Definition^[i].z > zMax)
        Then zMax := Obj.Object_Definition^[i].z;
    End;
  xSize := Abs(xMax) + Abs(xMin);
  ySize := Abs(yMax) + Abs(yMin);
  zSize := Abs(zMax) + Abs(zMin);
  If FitTexture
    Then Begin
      xScale := 256/xSize;
      yScale := 256/ySize;
      zScale := 256/zSize;
    End;

  For i := 0 to (Obj.Num_Verticies - 1) do
    Begin
      Case MappingAxis of
        XX : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        XY : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        XZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].x Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;
        YX : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        YY : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        YZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].y Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;
        ZX : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].x * xScale);
             End;
        ZY : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].y * yScale);
             End;
        ZZ : Begin
               tempu := (((Obj.Vertex_Normals^[i].z Div 2) + 127));
               tempv := 127 + Trunc(Obj.Object_Definition^[i].z * zScale);
             End;

      End;
      Clip(tempu, 0, 255);
      Clip(tempv, 0, 255);
      AddUVCoord(tempu, tempv, i);
    End;
End;

Procedure DoTexture;

Begin

  Case MappingStyle of
    PLANAR      : PlanarMapping;
    SPHERICAL   : SphericalMapping;
    CYLINDRICAL : CylindricalMapping;
  End;
End;

Begin
  ParseCommandLine;
  Writeln('Loading object.');
  InitObject;
  Writeln('Appling texture mapping algorithm.');
  DoTexture;
  Writeln('Writing new GVO file.');
  WriteGVOFile(OutFile);
End.